@@ -9,7 +9,7 @@
  * See the file "LICENSE.TERMS" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * $Id: fcgimisc.h,v 1.11 1996/10/29 20:34:42 mbrown Exp $
+ * $Id: fcgimisc.h,v 1.11.2.1 1997/01/18 15:51:53 snapper Exp $
  */
 
 #ifndef _FCGIMISC_H
@@ -18,16 +18,29 @@
 #include <stdio.h>
 #include <limits.h>
 
-#include <fcgi_config.h>
+#if defined (HAVE_CONFIG_H)
+#include <config.h>
+#endif
 
 #ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
 #endif
 #include <sys/types.h>
+#ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
+#endif
+#ifdef HAVE_WINSOCK_H
+#include <winsock.h>
+#endif
+#ifdef HAVE_NETINET_IN_H
 #include <netinet/in.h>
+#endif
+#ifdef HAVE_ARPA_INET_H
 #include <arpa/inet.h>
+#endif
+#ifdef HAVE_SYS_UN_H
 #include <sys/un.h>
+#endif
 
 /*
  * Where does this junk normally come from?
