@@ -214,7 +214,7 @@ pf_with_open_database (PFunArgs)
   int jump_again = 0;
 
   varname = mhtml_evaluate_string (get_positional_arg (vars, 0));
-  dbname = mhtml_evaluate_string (get_positional_arg (vars, 1));
+  dbname = mhtml_canonicalize_file_name_argument(get_positional_arg (vars, 1));
 
   if ((!empty_string_p (varname)) && (!empty_string_p (dbname)))
     {
@@ -263,8 +263,8 @@ pf_with_open_database (PFunArgs)
       unlock_database (dbname, &lock, &db);
     }
 
-  if (dbname) free (dbname);
-  if (varname) free (varname);
+  xfree (dbname);
+  xfree (varname);
   if (jump_again) longjmp (page_jmp_buffer, 1);
 }
 
@@ -1178,11 +1178,14 @@ pf_database_query (PFunArgs)
 	  if (!empty_string_p (sort_args))
 	    {
 	      char **sort_fields = (char **)NULL;
+	      char *sortorder;
 	      int sf_index = 0;
 	      int sf_slots = 0;
 	      char *temp_name;
 	      int offset = 0;
 
+	      sortorder = mhtml_evaluate_string(get_value (vars, "sortorder"));
+
 	      while (sort_args[offset])
 		{
 		  /* Skip whitespace and commas between field names. */
@@ -1226,26 +1229,48 @@ pf_database_query (PFunArgs)
 		    free (sort_fields[i]);
 		  free (sort_fields);
 		}
+
+	      if ((sortorder != (char *)NULL) &&
+		  ((strcasecmp (sortorder, "reverse") == 0) ||
+		   (strcasecmp (sortorder, "descending") == 0)))
+		{
+		  DBRecord **reversed = (DBRecord **)
+		    xmalloc ((1 + rec_index) * sizeof (DBRecord *));
+		  for (i = 0; i < rec_index; i++)
+		    reversed[(rec_index - 1) - i] = records[i];
+		  free (records);
+		  reversed[rec_index] = (DBRecord *)NULL;
+		  records = reversed;
+		}
 	    }
 
 	  /* If there is a format operator, evaluate it now. */
 	  if (format_expr != (char *)NULL)
 	    {
-	      int format_limit = -1;
 	      char *temp;
-	      char *fl;
+	      int format_limit = -1;
+	      int window_start = 0, window_length = 1000000;
 
-	      fl = mhtml_evaluate_string (get_value (vars, "format-limit"));
-	      if (!empty_string_p (fl))
+	      temp = mhtml_evaluate_string (get_value (vars, "window-start"));
+	      if (!empty_string_p (temp)) window_start = atoi (temp);
+	      xfree (temp);
+
+	      temp = mhtml_evaluate_string (get_value (vars, "window-length"));
+	      if (!empty_string_p (temp)) window_length = atoi (temp);
+	      xfree (temp);
+
+	      temp = mhtml_evaluate_string (get_value (vars, "format-limit"));
+	      if (!empty_string_p (temp))
 		{
-		  format_limit = atoi (fl);
+		  format_limit = atoi (temp);
 		  if (format_limit == 0) format_limit = -1;
 		}
+	      xfree (temp);
 
-	      if (fl != (char *)NULL) free (fl);
-
-	      for (i = 0; ((i < rec_index) &&
-			   ((format_limit < 0) || (i < format_limit))); i++)
+	      for (i = window_start;
+		   ((i < rec_index) && (i < window_length) &&
+		    ((format_limit < 0) || (i < format_limit)));
+		   i++)
 		{
 		  symbol_set_default_package (records[i]->contents);
 		  temp = mhtml_evaluate_string (format_expr);
