@@ -165,7 +165,7 @@ pf_with_open_database (PFunArgs)
   int jump_again = 0;
 
   varname = mhtml_evaluate_string (get_positional_arg (vars, 0));
-  dbname = mhtml_evaluate_string (get_positional_arg (vars, 1));
+  dbname = mhtml_canonicalize_file_name_argument(get_positional_arg (vars, 1));
 
   if ((!empty_string_p (varname)) && (!empty_string_p (dbname)))
     {
