@@ -481,7 +481,7 @@ forms_get_tag_value_in_package (Package *package, char *tag)
       char **values = symbol->values;
 
       /* Now find the appropriate variable in that realm. */
-      if (values != (char **)NULL && symbol->type == symtype_STRING)
+      if ((values != (char **)NULL) && (symbol->type == symtype_STRING))
 	{
 	  if (unarray)
 	    {
@@ -491,8 +491,9 @@ forms_get_tag_value_in_package (Package *package, char *tag)
 	      for (i = 0; values[i] != (char *)NULL; i++)
 		length += (2 + strlen (values[i]));
 
-	      value = (char *)xmalloc (length);
+	      value = (char *)xmalloc (1 + length);
 	      temp = value;
+	      *value = '\0';
 
 	      for (i = 0; values[i] != (char *)NULL; i++)
 		{
