@@ -349,6 +349,15 @@ get_cursor_ref (Package *vars)
 /*								   */
 /* *************************************************************** */
 
+/* CHAR type will be anything which is not explicitly numeric */ 
+#define CHARTYPE(ftype) ((! (gsql_field_type (ftype) == GSQL_NUMERIC)) \
+			 && (! (gsql_field_type (ftype) == GSQL_DECIMAL)) \
+			 && (! (gsql_field_type (ftype) == GSQL_INTEGER)) \
+			 && (! (gsql_field_type (ftype) == GSQL_SMALLINT)) \
+			 && (! (gsql_field_type (ftype) == GSQL_FLOAT)) \
+			 && (! (gsql_field_type (ftype) == GSQL_REAL)) \
+			 && (! (gsql_field_type (ftype) == GSQL_DOUBLE)))
+
 static int database_environment_level = 0;
 
 /* We pass an string which contains a numeric representation of the
@@ -545,11 +554,11 @@ get_dbref (Package *vars)
 static void
 pf_with_open_database (PFunArgs)
 {
-  char *varname  = mhtml_evaluate_string (get_positional_arg (vars, 0));
-  char *dsn      = mhtml_evaluate_string (get_value (vars, "DSN"));
-  char *dbname   = mhtml_evaluate_string (get_value (vars, "DATABASE"));
-  char *hostname = mhtml_evaluate_string (get_value (vars, "HOST"));
-  char *nolock   = mhtml_evaluate_string (get_value (vars, "NOLOCK"));
+  VOLATILE char *varname  = mhtml_evaluate_string (get_positional_arg (vars, 0));
+  VOLATILE char *dsn      = mhtml_evaluate_string (get_value (vars, "DSN"));
+  VOLATILE char *dbname   = mhtml_evaluate_string (get_value (vars, "DATABASE"));
+  VOLATILE char *hostname = mhtml_evaluate_string (get_value (vars, "HOST"));
+  VOLATILE char *nolock   = mhtml_evaluate_string (get_value (vars, "NOLOCK"));
   int jump_again = 0;
 
   /* No errors yet! */
@@ -577,7 +586,7 @@ pf_with_open_database (PFunArgs)
     {
       Database *db = make_database ();
       int lock;
-      int lock_p = 1;
+      volatile int lock_p = 1;
 
       if (!empty_string_p (nolock))
 	lock_p = 0;
@@ -587,7 +596,7 @@ pf_with_open_database (PFunArgs)
       if (db->connected == 1)
 	{
 	  char dbvalue[40];
-	  PAGE *body_code = page_copy_page (body);
+	  volatile PAGE *body_code = page_copy_page (body);
 	  int dbindex;
 
 	  /* Encode pointer to this database as an index into the table
@@ -604,14 +613,14 @@ pf_with_open_database (PFunArgs)
 	    database_environment_level++;
 
 	    if ((jump_again = setjmp (page_jmp_buffer)) == 0)
-	      page_process_page_internal (body_code);
+	      page_process_page_internal ((PAGE *)body_code);
 
 	    database_environment_level--;
 	    pagefunc_restore_environment (page_environ);
 
 	    if (jump_again && (body_code != (PAGE *)NULL))
 	      {
-		page_free_page (body_code);
+		page_free_page ((PAGE *)body_code);
 		body_code = (PAGE *)NULL;
 	      }
 	  }
@@ -624,7 +633,7 @@ pf_with_open_database (PFunArgs)
 		  *newstart = start + (body_code->bindex);
 		}
 
-	      page_free_page (body_code);
+	      page_free_page ((PAGE *)body_code);
 	    }
 	}
       else
@@ -656,6 +665,7 @@ static void
 pf_database_exec_sql (PFunArgs)
 {
   int dbindex;
+  gsql_result *query_result = (gsql_result *)NULL;
 
   if (database_environment_level != 0)
     {
@@ -663,13 +673,14 @@ pf_database_exec_sql (PFunArgs)
 
       if (db != (Database *)NULL)
 	{
-	  char *dbquery;
+	  char *dbquery = mhtml_evaluate_string (get_value (vars, "query"));
 
-	  dbquery = mhtml_evaluate_string (get_positional_arg (vars, 1));
+	  if (dbquery == (char *)NULL)
+	    dbquery = mhtml_evaluate_string (get_positional_arg (vars, 1));
 
 	  if (dbquery != (char *)NULL)
 	    {
-	      if (gsql_query (db, dbquery) == GSQL_ERROR)
+	      if (gsql_query (db, dbquery, 0) == GSQL_ERROR)
 		{
 		  gsql_save_error_message (db, GSQL_DEFAULT_ERRMSG);
 		}
@@ -678,13 +689,21 @@ pf_database_exec_sql (PFunArgs)
 		  bprintf_insert (page, start, "true");
 		  *newstart += 4;
 		}
+
+	      query_result = gsql_store_result (db);
+
+	      if (query_result != (gsql_result *) NULL)
+		{
+		  gsql_fetch_row (query_result);
+		  gsql_free_result (query_result);
+		}
+	      
 	      free (dbquery);
 	    }
 	}
     }
 }
 
-
 /* <database-exec-query dbvar query-string [cursor=cursor]>
    Executes an SQL query, within the scope of with-open-database.
 
@@ -704,13 +723,14 @@ pf_database_exec_query (PFunArgs)
 
       if (db != (Database *)NULL)
 	{
-	  char *dbquery;
+	  char *dbquery = mhtml_evaluate_string (get_value (vars, "query"));
 
-	  dbquery = mhtml_evaluate_string (get_positional_arg (vars, 1));
+	  if (dbquery == (char *)NULL)
+	    dbquery = mhtml_evaluate_string (get_positional_arg (vars, 1));
 
 	  if (dbquery != (char *)NULL)
 	    {
-	      if (gsql_query (db, dbquery) == GSQL_ERROR)
+	      if (gsql_query (db, dbquery, 0) == GSQL_ERROR)
 		{
 		  gsql_save_error_message (db, GSQL_DEFAULT_ERRMSG);
 		}
@@ -724,21 +744,44 @@ pf_database_exec_query (PFunArgs)
 
 		  query_result = gsql_store_result (db);
 
-		  /* Make a cursor for this result */
-		  cursor = make_cursor (db);
-		  cursor->result = query_result;
-
-		  /* Push the cursor on the databases's cursor stack. */
-		  cursor_index = istack_push (db->cursors, (void *) cursor);
-
-		  /* Make a token of db::cursor id and print it. */
-		  cursor_token = ((dbindex & 0xff) << 8)|(cursor_index & 0xff);
-
-		  sprintf (cursor_string, "%x", cursor_token);
-
+		  /* If the used passed in a valid cursor,
+		     we free its resources and re-use it. Otherwise we
+		     cons up a new one.
+		     */
 		  cursorvar =
 		    mhtml_evaluate_string (get_value (vars, "cursor"));
 
+		  cursor = lookup_cursor_ref (cursorvar);
+
+		  if (cursor == (DBCursor *)NULL)
+		    {
+		      /* Make a cursor for this result */
+		      cursor = make_cursor (db);
+		      cursor->result = query_result;
+
+		      /* Push the cursor on the databases's cursor stack. */
+		      cursor_index = istack_push
+			(db->cursors, (void *) cursor);
+
+		      /* Save the stack index of the cursor, so we can
+			 reference it in the stack if we need to. */
+		      cursor->index = cursor_index;
+		    }
+		  else
+		    {
+		      /* The cursor already exists, so free its resources
+			 (previously allocated db-specific HSTMT or cursors)
+			 and then store the newly allocate new ones.*/
+		      gsql_free_result (cursor->result);
+		      cursor->result = query_result;
+		    }
+
+		  /* Make a token of db::cursor id and print it. */
+		  cursor_token =
+		    ((dbindex & 0xff) << 8) | (cursor->index & 0xff);
+
+		  sprintf (cursor_string, "%x", cursor_token);
+
 		  if (!empty_string_p (cursorvar))
 		    {
 		      pagefunc_set_variable (cursorvar, cursor_string);
@@ -946,16 +989,12 @@ split_colnames (char *string, int *ncols)
   return (result);
 }
 
-static char *
-alist_of_field_info (gsql_field *field, Database *db);
-
+static char *alist_of_field_info (gsql_field *field, Database *db);
 
 /* <database-query-info cursor [result=varname]> 
 
    Returns an array of alists of information about the columns of the
-   given cursor's query result set.
-
- */
+   given cursor's query result set. */
 static void
 pf_database_query_info (PFunArgs)
 {
@@ -1099,9 +1138,9 @@ pf_database_next_record (PFunArgs)
 			break;
 		      else
 			{
-			  if (gsql_field_table (field)
-			      && (gsql_database_sql_prefix_tablenames (cursor->db)
-				  || !(empty_string_p (prefixtablenames))))
+			  if (gsql_field_table (field) &&
+			      (gsql_database_sql_prefix_tablenames (cursor->db)
+			       || !(empty_string_p (prefixtablenames))))
 			    {
 			      strcpy (buf, gsql_field_table (field));
 			      strcat (buf, ".");
@@ -1182,6 +1221,7 @@ pf_database_next_record (PFunArgs)
 	      free_array (fieldnames);
 	      xfree (pkgname);
 	      xfree (prefixtablenames);
+	      xfree (appendvars);
 	    }
 	}
     }
@@ -1242,7 +1282,6 @@ pf_database_set_options (PFunArgs)
 	    [tabletype=VAL]>
 
    Returns resulting array, or places it into VAR if supplied. */
-
 static void
 pf_database_tables (PFunArgs)
 {
@@ -1481,7 +1520,7 @@ package_from_row (gsql_result *query_result,
    If you want to collect 'keys', you need to specify a variable to
    put them into, and a key field name collect. Technically there is nothing
    which distinguishes this field from any other, it is just that we
-   need to speciy *some* field to collect the values of. */
+   need to specify *some* field to collect the values of. */
 typedef struct
 {
   char *key;
@@ -1503,14 +1542,17 @@ pf_database_query (PFunArgs)
       char *keyname = mhtml_evaluate_string (get_value (vars, "KEYNAME"));
       char *colnames = mhtml_evaluate_string (get_value (vars, "COLNAMES"));
       char *prefixtablenames =
-	mhtml_evaluate_string (get_value (vars, "prefixtablenames"));
+	mhtml_evaluate_string (get_value (vars, "PrefixTableNames"));
       char **fieldnames = (char **)NULL;
       gsql_result *query_result = (gsql_result *)NULL;
 
+      if (query == (char *)NULL)
+	query = mhtml_evaluate_string (get_positional_arg (vars, 1));
+
       /* Execute the query. */
       if (!empty_string_p (query))
 	{
-	  if (gsql_query (db, query) == GSQL_SUCCESS)
+	  if (gsql_query (db, query, 1) == GSQL_SUCCESS)
 	    {
 	      query_result = gsql_store_result (db);
 	      g_current_query_result = query_result;
@@ -1523,28 +1565,45 @@ pf_database_query (PFunArgs)
 	  char *expr = get_positional_arg (vars, 1);
 	  Package *save = CurrentPackage;
 	  DBRecord **records = (DBRecord **)NULL;
+	  int window_start = 0;
+	  int window_length = 9999999;
 	  int rec_slots = 0;
 	  int rec_index = 0;
-	  int search_limit = -1;
+	  int rec_count = 0;
 	  int ncolnames = 0;
-
-	  {
-	    char *search_limit_string =
-	      mhtml_evaluate_string (get_value (vars, "search-limit"));
-
-	    if (!empty_string_p (search_limit_string))
-	      {
-		search_limit = atoi (search_limit_string);
-		if (search_limit == 0) search_limit = -1;
-	      }
-
-	    xfree (search_limit_string);
-	  }
+	  int window_limit = window_length;
 
 	  if (!empty_string_p (colnames))
 	    fieldnames = split_colnames (colnames, &ncolnames);
 
-	  while ((search_limit < 0) || (rec_index < search_limit))
+	  /* Discard records until we get to the start of the window. */
+	  {
+	    char *temp;
+
+	    temp = mhtml_evaluate_string (get_value (vars, "window-start"));
+	    if (!empty_string_p (temp)) window_start = atoi (temp);
+	    xfree (temp);
+
+	    temp = mhtml_evaluate_string (get_value (vars, "window-length"));
+	    if (!empty_string_p (temp)) window_length = atoi (temp);
+	    xfree (temp);
+
+	    if (window_start < 0) window_start = 0;
+	    window_limit = window_start + window_length;
+
+	    while (rec_count < window_start)
+	      {
+		int status = gsql_fetch_row (query_result);
+
+		if ((status != GSQL_SUCCESS) &&
+		    (status != GSQL_SUCCESS_WITH_INFO))
+		  break;
+		else
+		  rec_count++;
+	      }
+	  }
+
+	  while (rec_count < window_limit)
 	    {
 	      int status = gsql_fetch_row (query_result);
 	      Package *db_fields;
@@ -1553,6 +1612,8 @@ pf_database_query (PFunArgs)
 	      if ((status != GSQL_SUCCESS) &&
 		  (status != GSQL_SUCCESS_WITH_INFO))
 		break;
+	      else
+		rec_count++;
 
 	      /* Otherwise, get the field data. */
 	      db_fields = package_from_row
@@ -1606,10 +1667,36 @@ pf_database_query (PFunArgs)
 		}
 	    }
 
+	  /* Count the rest of the records? */
+	  {
+	    char *countvar = mhtml_evaluate_string (get_value (vars, "count"));
+
+	    if (!empty_string_p (countvar))
+	      {
+		char num[50];
+
+		while (1)
+		  {
+		    int status = gsql_fetch_row (query_result);
+
+		    if ((status != GSQL_SUCCESS) &&
+			(status != GSQL_SUCCESS_WITH_INFO))
+		      break;
+		    else
+		      rec_count++;
+		  }
+
+		sprintf (num, "%d", rec_count);
+		pagefunc_set_variable (countvar, num);
+	      }
+
+	    xfree (countvar);
+	  }
+
 	  free_array (fieldnames);
 
-	  /* If there are any matching records, then sort, format,
-	     and/or return the keys. */
+	  /* If there are any matching records, then format, and/or
+	     return the keys. */
 	  if (rec_index != 0)
 	    {
 	      char *format_expr = get_value (vars, "format");
@@ -1621,8 +1708,8 @@ pf_database_query (PFunArgs)
 		  char *temp;
 		  char *fl;
 
-		  fl = mhtml_evaluate_string
-		    (get_value (vars, "format-limit"));
+		  fl =
+		    mhtml_evaluate_string (get_value (vars, "format-limit"));
 
 		  if (!empty_string_p (fl))
 		    {
@@ -1745,7 +1832,7 @@ bprintf_gsql_escape (BPRINTF_BUFFER *out, char *src, gsql_field *field,
   if (field != (gsql_field *)NULL)
     {
       fieldtype = gsql_field_type (field);
-      if (fieldtype == GSQL_CHAR)
+      if (CHARTYPE(field))
 	maxlength = gsql_field_length (field);
     }
 
@@ -1753,13 +1840,7 @@ bprintf_gsql_escape (BPRINTF_BUFFER *out, char *src, gsql_field *field,
     {
       /* If the src string is empty,
 	 SQL fields with numeric types get NULL values. */
-      if ((fieldtype == GSQL_NUMERIC) ||
-	  (fieldtype == GSQL_DECIMAL) ||
-	  (fieldtype == GSQL_INTEGER) ||
-	  (fieldtype == GSQL_SMALLINT) ||
-	  (fieldtype == GSQL_FLOAT) ||
-	  (fieldtype == GSQL_REAL) ||
-	  (fieldtype == GSQL_DOUBLE))
+      if (!(CHARTYPE(field)))
 	bprintf (out, "NULL");
       return;
     }
@@ -1769,8 +1850,14 @@ bprintf_gsql_escape (BPRINTF_BUFFER *out, char *src, gsql_field *field,
       i++;
       if ((maxlength != 0) && (i > maxlength)) return;
 
+      /* Oracle doesn't need parens quoted, but MSQL does */
+#if defined (COMPILING_MSQLFUNCS)
       if ((c == '\'') || (c == '(') || (c == ')'))
-	bprintf (out, escape_string);
+	bprintf (out, "%s", escape_string);
+#else
+      if (c == '\'')
+	bprintf (out, "%s", escape_string);
+#endif
 
       bprintf (out, "%c",c);
     }
@@ -1818,7 +1905,9 @@ lookup_fieldname (char *name, gsql_result *result)
    quotes around the data values (for char type). */
 
 #define QUOTE_KEYVAL_IF_NEEDED \
-	if (gsql_field_type (keyfield) == GSQL_CHAR) bprintf (query, "'")
+	if ((gsql_field_type (keyfield) == GSQL_CHAR) || \
+	    (gsql_field_type (keyfield) == GSQL_VARCHAR)) bprintf (query, "'")
+
 static void
 pf_database_save_record (PFunArgs)
 {
@@ -1846,8 +1935,8 @@ pf_database_save_record (PFunArgs)
       char *name;
       int position;
       int status;
-      gsql_field *field;
-      gsql_field *keyfield;
+      gsql_field *field = (gsql_field *)NULL;
+      gsql_field *keyfield = (gsql_field *)NULL;
       char escape_char = gsql_database_sql_escape_character (db);
       int truncate_p = gsql_database_sql_truncate_columns (db);
 
@@ -1938,13 +2027,13 @@ pf_database_save_record (PFunArgs)
 		  bprintf (query, ",");
 
 		  /* We only use single quotes around CHAR type data. */
-		  if (gsql_field_type (field) == GSQL_CHAR)
+		  if (CHARTYPE (field))
 		    bprintf (query, "'");
 
 		  bprintf_gsql_escape
 		    (query, value, field, escape_char, truncate_p);
 
-		  if  (gsql_field_type (field) == GSQL_CHAR)
+		  if (CHARTYPE (field))
 		    bprintf (query, "'");
 		}
 	    }
@@ -1952,12 +2041,18 @@ pf_database_save_record (PFunArgs)
 	  bprintf (query, ")");
 
 	  /* Execute the query. */
-	  status = gsql_query (db, query->buffer);
+	  if (debug_level)
+	    page_debug ("  Trying: <code>[%s]</code>", query->buffer);
+
+	  status = gsql_query (db, query->buffer, 0);
 	  bprintf_free_buffer (query);
 
 	  if (status == GSQL_SUCCESS)
+	    result = "true";
+	  else
 	    {
-	      result = "true";
+	      if (!empty_string_p (method))
+		gsql_save_error_message (db, GSQL_DEFAULT_ERRMSG);
 	    }
 	}
 
@@ -2001,11 +2096,11 @@ pf_database_save_record (PFunArgs)
 		  items_printed++;
 
 		  bprintf (query, "%s = ", gsql_field_name (field));
-		  if (gsql_field_type (field) == GSQL_CHAR)
+		  if (CHARTYPE(field))
 		    bprintf (query, "'");
 		  bprintf_gsql_escape
 		    (query, value, field, escape_char, truncate_p);
-		  if (gsql_field_type (field) == GSQL_CHAR)
+		  if (CHARTYPE(field))
 		    bprintf (query, "'");
 
 		}
@@ -2020,13 +2115,14 @@ pf_database_save_record (PFunArgs)
 	  QUOTE_KEYVAL_IF_NEEDED;
 
 	  /* Execute the query. */
-	  status = gsql_query (db, query->buffer);
+	  if (debug_level)
+	    page_debug ("  Trying: <code>[%s]</code>", query->buffer);
+
+	  status = gsql_query (db, query->buffer, 1);
 	  bprintf_free_buffer (query);
 
 	  if (status == GSQL_SUCCESS)
 	    result = "true";
-	  else
-	    gsql_save_error_message (db, GSQL_DEFAULT_ERRMSG);
 	}
     }
 
@@ -2110,7 +2206,10 @@ pf_database_load_record (PFunArgs)
 	  bprintf_gsql_escape (query, key, keyfield, escape_char, truncate_p);
 	  QUOTE_KEYVAL_IF_NEEDED;
 
-	  status = gsql_query (db, query->buffer);
+	  if (debug_level)
+	    page_debug ("  Trying: <code>[%s]</code>", query->buffer);
+
+	  status = gsql_query (db, query->buffer, 0);
 	  bprintf_free_buffer (query);
 	}
 
@@ -2235,7 +2334,10 @@ pf_database_delete_record (PFunArgs)
 	  bprintf_gsql_escape (query, key, keyfield, escape_char, truncate_p);
 	  QUOTE_KEYVAL_IF_NEEDED;
 
-	  status = gsql_query (db, query->buffer);
+	  if (debug_level)
+	    page_debug ("  Trying: <code>[%s]</code>", query->buffer);
+
+	  status = gsql_query (db, query->buffer, 0);
 	  bprintf_free_buffer (query);
 	}
 
@@ -2386,16 +2488,12 @@ pf_database_save_package (PFunArgs)
 		  bprintf (query, ",");
 
 		  /* We only use single quotes around CHAR type data. */
-		  if (gsql_field_type (field) == GSQL_CHAR)
+		  if (CHARTYPE(field))
 		    bprintf (query, "'");
 
 		  /* Print out the various values.  If there are none,
 		     print the empty string. */
-		  if (sym->values_index == 0)
-		    {
-		      bprintf (query, "''");
-		    }
-		  else
+		  if (sym->values_index != 0)
 		    {
 		      if (sym->values_index == 1)
 			bprintf_gsql_escape (query, sym->values[0], field,
@@ -2406,7 +2504,7 @@ pf_database_save_package (PFunArgs)
 			  BPRINTF_BUFFER *tmpbuf = bprintf_create_buffer ();
 
 			  for (j = 0; j < sym->values_index; j++)
-			    bprintf (tmpbuf, sym->values[j]);
+			    bprintf (tmpbuf, "%s", sym->values[j]);
 
 			  bprintf_gsql_escape (query, tmpbuf->buffer, field,
 					       escape_char, truncate_p);
@@ -2415,7 +2513,7 @@ pf_database_save_package (PFunArgs)
 			}
 		    }
 
-		  if  (gsql_field_type (field) == GSQL_CHAR)
+		  if (CHARTYPE(field))
 		    bprintf (query, "'");
 		}
 	    }
@@ -2423,11 +2521,19 @@ pf_database_save_package (PFunArgs)
 	  bprintf (query, ")");
 
 	  /* Execute the query. */
-	  status = gsql_query (db, query->buffer);
+	  if (debug_level)
+	    page_debug ("  Trying: <code>[%s]</code>", query->buffer);
+
+	  status = gsql_query (db, query->buffer, 0);
 	  bprintf_free_buffer (query);
 
 	  if (status == GSQL_SUCCESS)
 	    result = "true";
+	  else
+	    {
+	      if (empty_string_p (method))
+		gsql_save_error_message (db, GSQL_DEFAULT_ERRMSG);
+	    }
 	}
 
       if (!errors_found &&
@@ -2467,14 +2573,10 @@ pf_database_save_package (PFunArgs)
 
 		  bprintf (query, "%s = ", gsql_field_name (field));
 
-		  if (gsql_field_type (field) == GSQL_CHAR)
+		  if (CHARTYPE(field))
 		    bprintf (query, "'");
 
-		  if (sym->values_index == 0)
-		    {
-		      bprintf (query, "''");
-		    }
-		  else
+		  if (sym->values_index != 0)
 		    {
 		      if (sym->values_index == 1)
 			bprintf_gsql_escape (query, sym->values[0], field,
@@ -2487,7 +2589,7 @@ pf_database_save_package (PFunArgs)
 			  BPRINTF_BUFFER *tmpbuf = bprintf_create_buffer ();
 
 			  for (j = 0; j < sym->values_index; j++)
-			    bprintf (tmpbuf, sym->values[j]);
+			    bprintf (tmpbuf, "%s", sym->values[j]);
 
 			  bprintf_gsql_escape (query, tmpbuf->buffer, field,
 					       escape_char, truncate_p);
@@ -2496,7 +2598,7 @@ pf_database_save_package (PFunArgs)
 			}
 		    }
 
-		  if (gsql_field_type (field) == GSQL_CHAR)
+		  if (CHARTYPE(field))
 		    bprintf (query, "'");
 		}
 	    }
@@ -2509,7 +2611,10 @@ pf_database_save_package (PFunArgs)
 	  QUOTE_KEYVAL_IF_NEEDED;
 
 	  /* Execute the query. */
-	  status = gsql_query (db, query->buffer);
+	  if (debug_level)
+	    page_debug ("  Trying: <code>[%s]</code>", query->buffer);
+
+	  status = gsql_query (db, query->buffer, 0);
 	  bprintf_free_buffer (query);
 
 	  if (status == GSQL_SUCCESS)
@@ -2749,7 +2854,7 @@ pf_database_column_info (PFunArgs)
 	  tablename = mhtml_evaluate_string (get_positional_arg (vars, 1));
 	  fieldname = mhtml_evaluate_string (get_positional_arg (vars, 2));
 
-	  if ( !empty_string_p (tablename) && !empty_string_p (tablename) )
+	  if ((!empty_string_p (tablename)) && (!empty_string_p (fieldname)))
 	    {
 	      gsql_result *result = gsql_list_fields (db, tablename);
 
@@ -2880,21 +2985,25 @@ alist_from_row (gsql_result *query_result)
 
    Each alist contains database-specific information, including at least
    the NAME of the table.
-   ((name . "table_name"))
-
-*/
-
+   ((name . "table_name")) */
 static void
 pf_database_tables_info (PFunArgs)
 {
-  char *qualifier = mhtml_evaluate_string (get_value (vars, "tablequalifier"));
-  char *owner     = mhtml_evaluate_string (get_value (vars, "tableowner"));
-  char *name      = mhtml_evaluate_string (get_value (vars, "tablename"));
-  char *type      = mhtml_evaluate_string (get_value (vars, "tabletype"));
+  char *qualifier = mhtml_evaluate_string
+    (get_one_of (vars, "tablequalifier", "qualifier", (char *)NULL));
+  char *owner = mhtml_evaluate_string
+    (get_one_of (vars, "tableowner", "owner", (char *)NULL));
+  char *name = mhtml_evaluate_string
+    (get_one_of (vars, "tablename", "name", (char *)NULL));
+  char *type = mhtml_evaluate_string
+    (get_one_of (vars, "tabletype", "type", (char *)NULL));
 
   /* No errors yet! */
   gsql_clear_error_message ();
 
+  if (name == (char *)NULL)
+    name = mhtml_evaluate_string (get_positional_arg (vars, 1));
+
   if (database_environment_level != 0)
     {
       Database *db = get_dbref (vars);
@@ -2957,4 +3066,8 @@ pf_database_tables_info (PFunArgs)
 	    }
 	}
     }
+  xfree (qualifier);
+  xfree (owner);
+  xfree (name);
+  xfree (type);
 }
