@@ -134,6 +134,16 @@ pf_load_module (PFunArgs)
 	      char *no_init = mhtml_evaluate_string
 		(get_value (vars, "noinitialize"));
 
+	      /* Some operating systems lie.  For example, FreeBSD
+		 says that dlsym works.  But it apparently doesn't.
+		 Don't know why, can't say how.  It ALWAYS calls
+		 _init() though, so I guess we can just change all
+		 of our loadable libraries to have that function.
+		 Can you say "UGH?" */
+#if defined (__FreeBSD__)
+	      if (no_init == (char *)NULL)
+		no_init = strdup ("true");
+#endif
 	      if (no_init == (char *)NULL)
 		{
 		  VFUN *initfunc = (VFUN *)
@@ -159,7 +169,7 @@ pf_load_module (PFunArgs)
 		  symbol_add_value (sym, fqn);
 		}
 
-	      if (no_init) free (no_init);
+	      xfree (no_init);
 	    }
 	  else
 	    {
