@@ -28,62 +28,9 @@
    Santa Barbara, CA
    93101  */
 
-#define LANGUAGE_DEFINITIONS_FILE 1
+#include "language.h"
 
-#if defined (HAVE_CONFIG_H)
-#  include <config.h>
-#endif
-
-#include <stdio.h>
-#include <unistd.h>
-#include <stdlib.h>
-#include <stdarg.h>
-#include <string.h>
-#include <ctype.h>
-#include <regex.h>
-#include <setjmp.h>
-#include <sys/types.h>
-#if defined (Solaris)
-#  include <ucbinclude/sys/fcntl.h>
-#  include <ucbinclude/sys/file.h>
-#else
-#  include <sys/file.h>
-#endif /* !Solaris */
-#include <time.h>
-#include <math.h>
-
-#include <bprintf/bprintf.h>
-#include <xmalloc/xmalloc.h>
-#include <wisper/wisp.h>
-#include "database.h"
-#include "session_data.h"
-#include "pages.h"
-#include "parser.h"
-
-#define MHTML_STREAMS 1
-#define MHTML_DATABASE 1
-#define MHTML_CGI_EXEC 1
 #define MHTML_ARITHMETIC 1
-#define MHTML_BYTE_COMPILE 1
-
-#if defined (MHTML_STREAMS)
-#  include "streamfuncs.h"
-#endif /* MHTML_STREAMS */
-
-#include <locking.h>
-
-#if !defined (HAVE_SRANDOM)
-#  define srandom(seed) srand (seed)
-#  define random() rand()
-#endif
-
-#if defined (macintosh)
-extern char *strdup (const char *string);
-#  define os_open(name, flags, mode) open (name, flags)
-#else
-#  define os_open(name, flags, mode) open (name, flags, mode)
-#endif
-
 
 /************************************************************/
 /*							    */
@@ -111,6 +58,9 @@ static void pf_random (PFunArgs);
 /* <set-var name=value> */
 static void pf_set_variable (PFunArgs);
 
+/* <set-var-readonly name=value> */
+static void pf_set_variable_readonly (PFunArgs);
+
 /* <get-var name> --> value */
 static void pf_get_variable (PFunArgs);
 
@@ -168,6 +118,13 @@ static void pf_var_case (PFunArgs);
    Do replacements on %1..%9 and %body in BODY at invocation time. */
 static void pf_defmacro (PFunArgs);
 
+/* <defweakmacro macro-name> body </defmacro>
+   Store MACRO-NAME as a complex tag.
+   Do replacements on %1..%9 and %body in BODY at invocation time.
+   At invocation time, the closing tag does not necessarily have to be
+   present -- it is treated as a "subst" in that case. */
+static void pf_defweakmacro (PFunArgs);
+
 /* <defun fun-name> body </defun>
    Store FUN-NAME as a complex tag.
    By default, execution takes place in the local package with
@@ -240,9 +197,6 @@ static void pf_subst_in_var (PFunArgs);
 /* <subst-in-string string this that> Replace occurrences of THIS with THAT. */
 static void pf_subst_in_string (PFunArgs);
 
-/* <sort array-var [sort-fun] [caseless=true]> */
-static void pf_sort (PFunArgs);
-
 /* <date> --> Mon Feb 12 04:28:15 PST 1996 */
 static void pf_date (PFunArgs);
 
@@ -257,7 +211,7 @@ static void pf_symbol_info (PFunArgs);
    When action is "delete", returns the unmatched part STRING.
    When action is "startpos", returns the numeric offset of the start of
    the matched substring.
-   When action is "endtpos", returns the numeric offset of the end of the
+   When action is "endpos", returns the numeric offset of the end of the
    matched substring. */
 static void pf_match (PFunArgs);
 
@@ -295,6 +249,14 @@ static void pf_html_quote (PFunArgs);
 
 #endif /* DEPRECATED */
 
+/* <DEBUGGING-OUTPUT [arg]> 
+   -> Placeholder for output, or produce output immediately. */
+static void pf_debugging_output (PFunArgs);
+
+/* <SYSTEM-ERROR-OUTPUT [arg]> 
+   -> Placeholder for output, or produce output immediately. */
+static void pf_system_error_output (PFunArgs);
+
 /************************************************************/
 /*							    */
 /*			HTML "Helper" Functions		    */
@@ -312,6 +274,7 @@ PFunDesc pagefunc_table[] = {
 
   /* General use functions: variable manipulation and iteration operators. */
   { "SET-VAR",		0, 0, pf_set_variable },
+  { "SET-VAR-READONLY",	0, 0, pf_set_variable_readonly },
   { "GET-VAR",		0, 0, pf_get_variable },
   { "GET-VAR-ONCE",	0, 0, pf_get_variable_once },
   { "UNSET-VAR",	0, 0, pf_unset_variable },
@@ -337,6 +300,7 @@ PFunDesc pagefunc_table[] = {
   { "VAR-CASE",		0, 0, pf_var_case },
   { "WHILE",		1, 0, pf_while },
   { "DEFMACRO",		1, 0, pf_defmacro },
+  { "DEFWEAKMACRO",	1, 0, pf_defweakmacro },
   { "DEFINE-CONTAINER",	1, 0, pf_defmacro },
   { "DEFSUBST",		1, 0, pf_defsubst },
   { "DEFINE-TAG",	1, 0, pf_defsubst },
@@ -346,7 +310,6 @@ PFunDesc pagefunc_table[] = {
   { "FUNCTION-DEF",	0, 0, pf_function_def },
   { "DEBUGGING-ON",	0, 0, pf_debugging_on },
   { "PAGE-DEBUG",	0, 0, pf_page_debug },
-  { "SORT",		0, 0, pf_sort },
   { "DATE",		0, 0, pf_date },
 
   /* Random string operations. */
@@ -410,8 +373,10 @@ PFunDesc pagefunc_table[] = {
   { "RANDOMIZE",	0, 0, pf_randomize },
   { "RANDOM",		0, 0, pf_random },
 
-  { "TIME",		     0, 0, pf_time },
-  { "PID",		     0, 0, pf_pid },
+  { "TIME",		0, 0, pf_time },
+  { "PID",		0, 0, pf_pid },
+  { "DEBUGGING-OUTPUT",	0, 0, pf_debugging_output },
+  { "SYSTEM-ERROR-OUTPUT", 0, 0, pf_system_error_output },
 
   { (char *)NULL,	0, 0, (PFunHandler *)NULL }
 };
@@ -489,6 +454,25 @@ pf_defmacro (PFunArgs)
     free (subst_name);
 }
 
+static void
+pf_defweakmacro (PFunArgs)
+{
+  char *temp = get_positional_arg (vars, 0);
+  char *subst_name = mhtml_evaluate_string (temp);
+  char *subst_body = body ? body->buffer : "";
+
+  if (!empty_string_p (subst_name))
+    {
+      UserFunction *uf;
+      mhtml_add_user_function (user_MACRO, subst_name, subst_body, vars);
+      uf = mhtml_find_user_function (subst_name);
+      uf->flags |= user_WEAK_MACRO;
+    }
+
+  if (subst_name && subst_name != temp)
+    free (subst_name);
+}
+
 static void
 pf_defun (PFunArgs)
 {
@@ -652,8 +636,10 @@ pf_symbol_info (PFunArgs)
 }
 
 static void
-pf_set_variable (PFunArgs)
+generic_set_variable (Package *vars, int debug_level, int readonly_p)
 {
+  char *func = readonly_p ? "set-var-readonly" : "set-var";
+
   if (vars)
     {
       char **names = get_vars_names (vars);
@@ -675,15 +661,15 @@ pf_set_variable (PFunArgs)
 	      if (debug_level >= 5)
 		{
 		  if (value)
-		    page_debug ("<set-var \"%s\"=\"%s\">", sym_name, value);
+		    page_debug ("<%s \"%s\"=\"%s\">", func, sym_name, value);
 		  else
-		    page_debug ("<set-var \"%s\">");
+		    page_debug ("<%s \"%s\">", func, sym_name);
 		}
 
 	      if (value == (char *)NULL)
 		{
 		  if (debug_level)
-		    page_debug ("<set-var %s ...> missing `='", sym_name);
+		    page_debug ("<%s %s ...> missing `='", func, sym_name);
 		}
 	      else
 		{
@@ -692,11 +678,16 @@ pf_set_variable (PFunArgs)
 		}
 
 	      if (debug_level >= 6)
-		page_debug ("--> <set-var \"%s\"=\"%s\">",
-			    name ? name : "", value ? value : "");
+		page_debug ("--> <%s \"%s\"=\"%s\">",
+			    func, name ? name : "", value ? value : "");
 
 	      if (name)
-		pagefunc_set_variable (name, value);
+		{
+		  if (readonly_p)
+		    pagefunc_set_variable_readonly (name, value);
+		  else
+		    pagefunc_set_variable (name, value);
+		}
 
 	      if (free_value) free (value);
 	      if (name != sym_name) free (name);
@@ -705,6 +696,18 @@ pf_set_variable (PFunArgs)
     }
 }
 
+static void
+pf_set_variable (PFunArgs)
+{
+  generic_set_variable (vars, debug_level, 0);
+}
+
+static void
+pf_set_variable_readonly (PFunArgs)
+{
+  generic_set_variable (vars, debug_level, 1);
+}
+
 static void
 pf_get_variable (PFunArgs)
 {
@@ -791,12 +794,8 @@ pf_unset_variable (PFunArgs)
 
   for (i = 0; (name = get_positional_arg (vars, i)) != (char *)NULL; i++)
     {
-      char *varname = name;
-      Symbol *sym;
-
-      varname = mhtml_evaluate_string (name);
-
-      sym = symbol_lookup (varname);
+      char *varname = mhtml_evaluate_string (name);
+      Symbol *sym = varname ?  symbol_lookup (varname) : (Symbol *)NULL;
 
       if (sym)
 	{
@@ -816,13 +815,14 @@ pf_unset_variable (PFunArgs)
 
 	      sym->values_index = 0;
 	    }
-	  else
-	    sym = symbol_remove (varname);
-
-	  if (sym) symbol_free (sym);
+	  else if (!symbol_get_flag (sym, sym_READONLY))
+	    {
+	      sym = symbol_remove (varname);
+	      if (sym) symbol_free (sym);
+	    }
 	}
 
-      if (varname != name) free (varname);
+      xfree (varname);
     }
 }
 
@@ -835,7 +835,7 @@ pf_variable_set_p (PFunArgs)
   if (!empty_string_p (arg) && (symbol_lookup (arg) != (Symbol *)NULL))
     set_p++;
 
-  if (arg != (char *)NULL) free (arg);
+  xfree (arg);
 
   if (set_p)
     {
@@ -960,52 +960,31 @@ pf_package_delete (PFunArgs)
 static void
 pf_if (PFunArgs)
 {
-  char *test_clause;
-  char *then_clause;
-  char *else_clause;
+  char *test_clause = mhtml_evaluate_string (get_positional_arg (vars, 0));
+  char *then_clause = get_positional_arg (vars, 1);
+  char *else_clause = get_positional_arg (vars, 2);
   char *consequence;
 
-  test_clause = get_positional_arg (vars, 0);
-  then_clause = get_positional_arg (vars, 1);
-  else_clause = get_positional_arg (vars, 2);
-
-  test_clause = mhtml_evaluate_string (test_clause);
-
   if (!empty_string_p (test_clause))
     consequence = then_clause;
   else
     consequence = else_clause;
 
-  if (!empty_string_p (consequence))
-    {
-      consequence = mhtml_evaluate_string (consequence);
-      if (consequence != (char *)NULL)
-	{
-	  bprintf_insert (page, start, "%s", consequence);
-	  free (consequence);
-	}
-    }
+  if (consequence != (char *)NULL)
+    bprintf_insert (page, start, "%s", consequence);
 
-  if (test_clause) free (test_clause);
+  xfree (test_clause);
 }
 
 static void
 pf_ifeq (PFunArgs)
 {
-  char *left_clause;
-  char *right_clause;
-  char *then_clause;
-  char *else_clause;
-  char *consequence;
+  char *left_clause = mhtml_evaluate_string (get_positional_arg (vars, 0));
+  char *right_clause = mhtml_evaluate_string (get_positional_arg (vars, 1));
+  char *then_clause = get_positional_arg (vars, 2);
+  char *else_clause = get_positional_arg (vars, 3);
   int caseless_p = var_present_p (vars, "CASELESS");
-
-  left_clause = get_positional_arg (vars, 0);
-  right_clause = get_positional_arg (vars, 1);
-  then_clause = get_positional_arg (vars, 2);
-  else_clause = get_positional_arg (vars, 3);
-
-  left_clause = mhtml_evaluate_string (left_clause);
-  right_clause = mhtml_evaluate_string (right_clause);
+  char *consequence;
 
   if (((empty_string_p (left_clause)) && (empty_string_p (right_clause))) ||
       ((left_clause && right_clause) &&
@@ -1015,19 +994,11 @@ pf_ifeq (PFunArgs)
   else
     consequence = else_clause;
 
-  if (consequence)
-    {
-      consequence = mhtml_evaluate_string (consequence);
+  if (consequence != (char *)NULL)
+    bprintf_insert (page, start, "%s", consequence);
 
-      if (consequence)
-	{
-	  bprintf_insert (page, start, "%s", consequence);
-	  free (consequence);
-	}
-    }
-
-  if (left_clause) free (left_clause);
-  if (right_clause) free (right_clause);
+  xfree (left_clause);
+  xfree (right_clause);
 }
 
 #if defined (MHTML_ARITHMETIC)
@@ -1130,7 +1101,7 @@ arithmetic_operate (int op, char *arg1, char *arg2)
 {
   double val1 = arg1 ? strtod (arg1, (char **)NULL) : 0.0;
   double val2 = arg2 ? strtod (arg2, (char **)NULL) : 0.0;
-  static char resbuf[40];
+  static char resbuf[60];
   static int orig_mhtml_decimal_notify = 0;
   char *result = resbuf;
 
@@ -1211,7 +1182,7 @@ arithmetic_operate (int op, char *arg1, char *arg2)
 	if (mhtml_decimal_notify)
 	  sprintf (result, "%.*f", mhtml_decimal_places, arith_result);
 	else if (!dot_present /* || (arith_result == (int)arith_result) */)
-	  sprintf (result, "%d", (int) arith_result);
+	  sprintf (result, "%ld", (long int)arith_result);
 	else
 	  sprintf (result, "%.*f", mhtml_decimal_places, arith_result);
       }
@@ -1254,11 +1225,14 @@ arithmetic_process (int op, PFunArgs)
       char *rarg1 = (char *)NULL;
       char *rarg2 = (char *)NULL;
 
-      if (strchr (arg1, '<'))
+      if ((strchr (arg1, '<') != (char *)NULL) ||
+	  ((LEFT_BRACKET != '<') &&
+	   (strchr (arg1, LEFT_BRACKET) != (char *)NULL)))
 	arg1 = mhtml_evaluate_string (arg1);
 
-
-      if (strchr (arg2, '<'))
+      if ((strchr (arg2, '<') != (char *)NULL) ||
+	  ((LEFT_BRACKET != '<') &&
+	   (strchr (arg2, LEFT_BRACKET) != (char *)NULL)))
 	arg2 = mhtml_evaluate_string (arg2);
 
       rarg1 = arg1;
@@ -1323,8 +1297,16 @@ static void pf_add (PFunArgs)
 
 static void pf_sub (PFunArgs)
 {
-  arithmetic_process
-    (pf_SUB, page, body, vars, start, end, newstart, debug_level);
+  if ((body == (PAGE *)NULL) || (empty_string_p (body->buffer)))
+    {
+      bprintf_insert (page, start, "<SUB>");
+      *newstart += 4;
+    }
+  else
+    {
+      arithmetic_process
+	(pf_SUB, page, body, vars, start, end, newstart, debug_level);
+    }
 }
 
 static void pf_mul (PFunArgs)
@@ -1335,8 +1317,18 @@ static void pf_mul (PFunArgs)
 
 static void pf_div (PFunArgs)
 {
-  arithmetic_process
-    (pf_DIV, page, body, vars, start, end, newstart, debug_level);
+  /* If it only contains assigned attributes, it is meant for the browser,
+     not us. */
+  if (get_positional_arg (vars, 0) == (char *)NULL)
+    {
+      bprintf_insert (page, start, "<DIV %s>", mhtml_funargs (vars));
+      *newstart += 4;
+    }
+  else
+    {
+      arithmetic_process
+	(pf_DIV, page, body, vars, start, end, newstart, debug_level);
+    }
 }
 
 static void pf_mod (PFunArgs)
@@ -1387,7 +1379,7 @@ pf_random (PFunArgs)
 
   bprintf_insert (page, start, "%d", result);
 }
-
+  
 static void
 pf_when (PFunArgs)
 {
@@ -1474,7 +1466,6 @@ static void
 pf_while (PFunArgs)
 {
   char *test = get_positional_arg (vars, 0);
-  int verbatim_p = var_present_p (vars, "VERBATIM");
   int iteration_count = 0;
   char *limit_string = pagefunc_get_variable ("mhtml::iteration-limit");
   int limit = limit_string ? atoi (limit_string) : PAGE_ITERATOR_MAX_COUNT;
@@ -1498,8 +1489,7 @@ pf_while (PFunArgs)
 	  bprintf_insert (page, start, "%s", code->buffer);
 	  start += (code->bindex);
 
-	  if (verbatim_p)
-	    *newstart = start;
+	  *newstart = start;
 	}
 
       if (code != (PAGE *)NULL)
@@ -1855,6 +1845,7 @@ pf_downcase (PFunArgs)
 	  value[i] = tolower (value[i]);
 
       bprintf_insert (page, start, "%s", value);
+      *newstart += i;
       free (value);
     }
 }
@@ -1873,6 +1864,7 @@ pf_upcase (PFunArgs)
 	  value[i] = toupper (value[i]);
 
       bprintf_insert (page, start, "%s", value);
+      *newstart += i;
       free (value);
     }
 }
@@ -1909,6 +1901,7 @@ pf_capitalize (PFunArgs)
 	}
 
       bprintf_insert (page, start, "%s", value);
+      *newstart += i;
       free (value);
     }
 }
@@ -1959,10 +1952,14 @@ static void
 pf_word_wrap (PFunArgs)
 {
   char *width_spec = mhtml_evaluate_string (get_value (vars, "width"));
+  char *indent_spec = mhtml_evaluate_string (get_value (vars, "indent"));
   int width = (width_spec != (char *)NULL) ? atoi (width_spec) : 60;
+  int indent = (indent_spec != (char *)NULL) ? atoi (indent_spec) : 0;
   char *text = mhtml_evaluate_string (get_positional_arg (vars, 0));
 
   if (width == 0) width = 60;
+  if (indent > width) indent = width;
+  if (indent < 0) indent = 0;
 
   if (!empty_string_p (text))
     {
@@ -1971,12 +1968,41 @@ pf_word_wrap (PFunArgs)
       bprintf (temp, "%s", text);
       bprintf_word_wrap (temp, width);
 
+      if (indent)
+	{
+	  register int i;
+	  char *indent_string = (char *)xmalloc (1 + indent);
+
+	  for (i = 0; i < indent; i++) indent_string[i] = ' ';
+	  indent_string[i] = '\0';
+
+	  /* Kill leading indentation on the first line. */
+	  for (i = 0; (i < temp->bindex) && (whitespace (temp->buffer[i])); i++);
+	  if (i) bprintf_delete_range (temp, 0, i);
+
+	  /* Indent the first line. */
+	  bprintf_insert (temp, 0, "%s", indent_string);
+
+	  /* Now do the rest. */
+	  while (i < temp->bindex)
+	    {
+	      if (temp->buffer[i] == '\n')
+		{
+		  bprintf_insert (temp, i + 1, "%s", indent_string);
+		  i += indent;
+		}
+	      i++;
+	    }
+	}
+	  
       bprintf_insert (page, start, "%s", temp->buffer);
+      *newstart += temp->bindex;
       bprintf_free_buffer (temp);
     }
 
-  if (text != (char *)NULL) free (text);
-  if (width_spec != (char *)NULL) free (width_spec);
+  xfree (text);
+  xfree (width_spec);
+  xfree (indent_spec);
 }
 
 static void
@@ -2067,9 +2093,9 @@ pf_plain_text (PFunArgs)
 
   /* Insert the modified body. */
   bprintf_insert (page, start, "%s", body->buffer + (nobr ? 3 : 0));
-  if (nobr) free (nobr);
-  if (nolower) free (nolower);
-  if (first_char) free (first_char);
+  xfree (nobr);
+  xfree (nolower);
+  xfree (first_char);
 }
 
 static int include_recursive_calls = 0;
@@ -2182,7 +2208,6 @@ pf_replace_page (PFunArgs)
   int newpage_start = 0;
 
   pf_include (newpage, body, vars, 0, 0, &newpage_start, debug_level);
-  /* ThePage = newpage; */
   page_process_page_internal (newpage);
   page_return_this_page (newpage);
 }
@@ -2451,7 +2476,6 @@ pf_pad (PFunArgs)
 static void
 pf_subst_in_page (PFunArgs)
 {
-  int changed = 0;
   int arg = 0, done = 0;
 
   while (!done)
@@ -2469,14 +2493,13 @@ pf_subst_in_page (PFunArgs)
 	  that = mhtml_evaluate_string (that);
 
 	  if (this)
-	    changed += page_subst_in_page (page, this, that);
-
+	    page_subst_in_page_pivot (page, this, that, &start);
 
 	  if (this) free (this);
 	  if (that) free (that);
 	}
     }
-  *newstart = start + changed;
+  *newstart = start;
 }
 
 static char *
@@ -2705,8 +2728,7 @@ pf_redirect (PFunArgs)
       if (!empty_string_p (sexp))
 	arg =  mhtml_evaluate_string (sexp);
 
-      if (sexp != (char *)NULL)
-	free (sexp);
+      xfree (sexp);
     }
 
   new_page = page_create_page ();
@@ -2729,12 +2751,33 @@ pf_redirect (PFunArgs)
 	  BPRINTF_BUFFER *newarg = bprintf_create_buffer ();
 	  char *temp = pagefunc_get_variable ("mhtml::relative-prefix");
 
+	  /* We can't do a complete canonicalization of the name here,
+	     because there isn't any way to tell where the PATH_INFO
+	     part of the name begins, if any.  But, we can strip double
+	     dots which appear at the start of the name. */
+	  if (temp != (char *)NULL)
+	    {
+	      temp = strdup (temp);
+
+	      while (strncmp (arg + i, "../", 3) == 0)
+		{
+		  char *slash = strrchr (temp, '/');
+
+		  if (slash != (char *)NULL)
+		    {
+		      *slash = '\0';
+		      i += 3;
+		    }
+		}
+	    }
+
 	  bprintf (newarg, "%s", pagefunc_get_variable ("mhtml::http-prefix"));
 	  if (arg[i] != '/')
 	    bprintf (newarg, "%s/%s", temp ? temp : "", arg + i);
 	  else
 	    bprintf (newarg, "%s", arg + i);
 
+	  xfree (temp);
 	  free (arg);
 	  arg = newarg->buffer;
 	  free (newarg);
@@ -2771,6 +2814,12 @@ pf_cgi_encode (PFunArgs)
       char *result = (char *)NULL;
       Package *cgivars = symbol_get_package ((char *)NULL);
       Symbol **symbols = (Symbol **)NULL;
+      int save_case_p = 0;
+
+      {
+	char *temp = mhtml_evaluate_string (get_value (vars, "preserve-case"));
+	if (!empty_string_p (temp)) save_case_p++;
+      }
 
       for (i = 0; (name = names[i]) != (char *)NULL; i++)
 	{
@@ -2784,6 +2833,9 @@ pf_cgi_encode (PFunArgs)
 		  register int j;
 		  Symbol *newsym = symbol_intern_in_package (cgivars, name);
 
+		  if (save_case_p)
+		    newsym->preserved_name = strdup (name);
+
 		  for (j = 0; j < sym->values_index; j++)
 		    symbol_add_value (newsym, sym->values[j]);
 		}
@@ -3006,11 +3058,8 @@ pf_server_push (PFunArgs)
 
 	      called_yet++;
 
-#if defined (MHTML_STREAMS)
 	      output_fd = mhtml_stdout_fileno;
-#else
-	      output_fd = fileno (stdout);
-#endif
+
 	      if (nph)
 		bprintf (pushed, "HTTP/1.0 200\n");
 
@@ -3038,99 +3087,99 @@ pf_server_push (PFunArgs)
     }
 }
 
-static char *mhtml_sort_function_name = (char *)NULL;
-static int mhtml_sort_is_caseless = 0;
-static int mhtml_sort_is_descending = 0;
-
-static int
-sort_with_function (const void *item1, const void *item2)
-{
-  char *string1 = *(char **)item1;
-  char *string2 = *(char **)item2;
-  int should_free = 0;
-  int result;
-
-  if (mhtml_sort_function_name)
-    {
-      PAGE *page = page_create_page ();
-
-      bprintf (page, "<%s %s>", mhtml_sort_function_name, string1);
-      string1 = mhtml_evaluate_string (page->buffer);
-      page->bindex = 0;
-      bprintf (page, "<%s %s>", mhtml_sort_function_name, string2);
-      page->buffer[page->bindex] = '\0';
-      string2 = mhtml_evaluate_string (page->buffer);
-
-      page_free_page (page);
-      should_free++;
-    }
-
-  if (string1 && !string2)
-    result = 1;
-  else if (string2 && !string1)
-    result = -1;
-  else if (!string1 && !string2)
-    result = 0;
-  else if (mhtml_sort_is_caseless)
-    result = strcasecmp (string1, string2);
-  else
-    result = strcmp (string1, string2);
-
-  if (should_free && string1) free (string1);
-  if (should_free && string2) free (string2);
-
-  if (result && mhtml_sort_is_descending)
-    result = -result;
-
-  return (result);
-}
-
-static void
-pf_sort (PFunArgs)
-{
-  char *sortvar = mhtml_evaluate_string (get_positional_arg (vars, 0));
-
-  if (!empty_string_p (sortvar))
-    {
-      Symbol *sym = symbol_lookup (sortvar);
-
-      /* If there is anything to sort... */
-      if ((sym != (Symbol *)NULL) && (sym->values_index != 0))
-	{
-	  char *sorter = mhtml_evaluate_string (get_positional_arg (vars, 1));
-	  int caseless = (get_value (vars, "caseless") != (char *)NULL);
-	  int descending = (get_value (vars, "descending") != (char *)NULL);
-
-	  mhtml_sort_function_name = (char *)NULL;
-
-	  if (!empty_string_p (sorter))
-	    mhtml_sort_function_name = sorter;
-
-	  mhtml_sort_is_caseless = caseless;
-	  mhtml_sort_is_descending = descending;
-
-	  qsort ((void *)sym->values, sym->values_index, sizeof (char *),
-		 sort_with_function);
-
-	  if (sorter) free (sorter);
-	}
-    }
-
-  if (sortvar) free (sortvar);
-}
-
 static void
 pf_date (PFunArgs)
 {
-  /* We can handle GMT and friends later. */
   char *tstring = mhtml_evaluate_string (get_positional_arg (vars, 0));
+  char *gmt = get_value (vars, "gmt");
   time_t ticks = tstring ? (time_t)atol (tstring) : (time_t)time ((time_t *)0);
   char *time_string = ctime (&ticks);
   char *temp = strchr (time_string, '\n');
 
   if (temp) *temp = '\0';
 
+  if (gmt != (char *)NULL)
+    time_string = http_date_format ((long) ticks);
+
   bprintf_insert (page, start, "%s", time_string);
   *newstart += strlen (time_string);
   if (tstring) free (tstring);
 }
+
+static void
+pf_debugging_output (PFunArgs)
+{
+  register int i = 0;
+  char *text;
+  static char *token_name = "<DEBUGGING-OUTPUT>";
+
+  while ((text = get_positional_arg (vars, i)) != (char *)NULL)
+    {
+      char *arg = mhtml_evaluate_string (text);
+
+      if (!empty_string_p (arg))
+	{
+	  if (strcasecmp (arg, "clear") == 0)
+	    {
+	      page_debug_clear ();
+	    }
+	  else if (strcasecmp (arg, "retrieve") == 0)
+	    {
+	      char *contents = page_debug_buffer ();
+
+	      if (contents != (char *)NULL)
+		{
+		  bprintf_insert (page, start, "%s", contents);
+		  *newstart = start + strlen (contents) - 1;
+		}
+	    }
+	}
+      xfree (arg);
+      i++;
+    }
+
+  if (i == 0)
+    {
+      bprintf_insert (page, start, token_name);
+      *newstart = start + strlen (token_name) - 1;
+    }
+}
+
+static void
+pf_system_error_output (PFunArgs)
+{
+  register int i = 0;
+  char *text;
+  static char *token_name = "<SYSTEM-ERROR-OUTPUT>";
+
+  while ((text = get_positional_arg (vars, i)) != (char *)NULL)
+    {
+      char *arg = mhtml_evaluate_string (text);
+
+      if (!empty_string_p (arg))
+	{
+	  if (strcasecmp (arg, "clear") == 0)
+	    {
+	      page_syserr_clear ();
+	    }
+	  else if (strcasecmp (arg, "retrieve") == 0)
+	    {
+	      char *contents = page_syserr_buffer ();
+
+	      if (contents != (char *)NULL)
+		{
+		  bprintf_insert (page, start, "%s", contents);
+		  *newstart = start + strlen (contents) - 1;
+		}
+	    }
+	}
+      xfree (arg);
+      i++;
+    }
+
+  if (i == 0)
+    {
+      bprintf_insert (page, start, token_name);
+      *newstart = start + strlen (token_name) - 1;
+    }
+}
