@@ -80,12 +80,14 @@
 #include "mac_port.h"
 #endif
 
+extern char LEFT_BRACKET, RIGHT_BRACKET;
+
 #undef whitespace
 #define whitespace(c) \
      (((c) == ' ') || ((c) == '\t') || ((c) == '\r') || ((c) == '\n'))
 
 #define form_tag_ender(c) (whitespace (c) || (c == '>') || (c == '"'))
-#define tag_ender(c)	  (whitespace (c) || (c == '>'))
+#define tag_ender(c) (whitespace (c) || (c == '>') || (c == RIGHT_BRACKET))
 
 
 #if defined (macintosh)
@@ -235,14 +237,18 @@ page_read_template (char *passed_filename)
 
 /* Do some substitutions in PAGE; that is, substitute THIS WITH_THAT in PAGE.
    The substitutions happen in place, and the return value is the number of
-   characters added or deleted.  THIS is a regular expression. */
+   characters added or deleted before PIVOT.  THIS is a regular expression.
+   PIVOT is a pointer to an integer which is adjusted depending on the number
+   of characters added or deleted before it.  An NULL pointer means skip this
+   calculation. */
 int
-page_subst_in_page (PAGE *page, char *this, char *with_that)
+page_subst_in_page_pivot (PAGE *page, char *this, char *with_that, int *pivot)
 {
   regex_t re;
   regmatch_t offsets[10];
   int matched, search_point, done;
   int len = 0;
+  int point = (pivot ? *pivot : 0);
   int initial_characters = page->bindex;
   static int *replacement_offsets = (int *)NULL;
   static int ro_slots = 0;
@@ -299,6 +305,11 @@ page_subst_in_page (PAGE *page, char *this, char *with_that)
 
 	  if (ro_index == 0)
 	    {
+	      if (end < point)
+		point = point - (end - start) + len;
+	      else if (start < point)
+		point = start + len;
+
 	      bprintf_delete_range (page, start, end);
 	      bprintf_insert_text (page, start, with_that);
 
@@ -313,7 +324,7 @@ page_subst_in_page (PAGE *page, char *this, char *with_that)
 
 	      for (i = 0; i < ro_index; i++)
 		{
-		  int b = i == 0 ? 0 : replacement_offsets[i - 1] + 2;
+		  int b = ((i == 0) ? 0 : replacement_offsets[i - 1] + 2);
 		  int e = replacement_offsets[i];
 
 		  wt[e] = '\0';
@@ -336,6 +347,11 @@ page_subst_in_page (PAGE *page, char *this, char *with_that)
 
 	      free (wt);
 
+	      if (end < point)
+		point = point - (end - start) + buffer->bindex;
+	      else if (start < point)
+		point = start + buffer->bindex;
+
 	      bprintf_delete_range (page, start, end);
 	      bprintf_insert_text (page, start, buffer->buffer);
 	      search_point = start + buffer->bindex;
@@ -348,9 +364,21 @@ page_subst_in_page (PAGE *page, char *this, char *with_that)
 
   regfree (&re);
 
+  if (pivot != (int *)NULL)
+    *pivot = point;
+
   return (page->bindex - initial_characters);
 }
 
+/* Do some substitutions in PAGE; that is, substitute THIS WITH_THAT in PAGE.
+   The substitutions happen in place, and the return value is the number of
+   characters added or deleted.  THIS is a regular expression. */
+int
+page_subst_in_page (PAGE *page, char *this, char *with_that)
+{
+  return (page_subst_in_page_pivot (page, this, with_that, (int *)NULL));
+}
+
 /* Return the offset of STRING in PAGE.  Start the search at START.
    The absolute index is returned, not the offset from START, so
    the value is suitable for resubmission to this function as START value.
@@ -908,7 +936,7 @@ page_find_tag_start (PAGE *page, char *tag, int start)
 
   for (i = start; i < limit; i++)
     {
-      if (page->buffer[i] == '<')
+      if ((page->buffer[i] == '<') || (page->buffer[i] == LEFT_BRACKET))
 	{
 	  c = page->buffer[i + 1 + len];
 
@@ -933,7 +961,8 @@ page_find_tag_end (PAGE *page, int start)
 {
   int end = -1;
 
-  if ((start < page->bindex) && (page->buffer[start] == '<'))
+  if ((start < page->bindex) &&
+      ((page->buffer[start] == '<') || (page->buffer[start] == LEFT_BRACKET)))
     {
       register int i;
       int quote_depth = 0;
@@ -950,6 +979,9 @@ page_find_tag_end (PAGE *page, int start)
 	      continue;
 	    }
 
+	  if (c == LEFT_BRACKET) c = '<';
+	  if (c == RIGHT_BRACKET) c = '>';
+
 	  switch (c)
 	    {
 	    case '<':
@@ -1024,12 +1056,15 @@ page_complex_tag_bounds (PAGE *page, char *tag, int *startp, int *endp)
 
   /* Try to optimize... */
   if (((open_start + tlen + 1) < page->bindex) &&
-      (page->buffer[open_start] == '<') &&
+      ((page->buffer[open_start] == '<') ||
+       (page->buffer[open_start] == LEFT_BRACKET)) &&
       (strncasecmp (tag, page->buffer + open_start + 1, tlen) == 0) &&
       ((whitespace (page->buffer[open_start + tlen + 1]) ||
-	page->buffer[open_start + tlen + 1] == '>')))
+	((page->buffer[open_start + tlen + 1] == '>') ||
+	 (page->buffer[open_start + tlen + 1] == RIGHT_BRACKET)))))
     {
-      if (page->buffer[open_start + tlen + 1] == '>')
+      if ((page->buffer[open_start + tlen + 1] == '>') ||
+	  (page->buffer[open_start + tlen + 1] == RIGHT_BRACKET))
 	open_end = open_start + tlen + 1;
       else
 	open_end = page_find_tag_end (page, open_start);
@@ -1103,6 +1138,70 @@ page_complex_tag_bounds (PAGE *page, char *tag, int *startp, int *endp)
   return (result);
 }
 
+/* Find out if <indicator> belongs to the complex <tag>, and return 3 values:
+      1) The output of the function is non-zero if INDICATOR was found,
+         It is 1 if the indicator was found and it was ours, -1 if the
+	 indicator was found and it wasn't ours, and 0 if the indicator
+	 wasn't found at all.
+      2) The value of STARTP gets the offset of INDICATOR,
+      3) the value of ENDP gets the offset of the end of INDICATOR. */
+int
+page_indicator_owned_by (PAGE *page, char *indicator, char *tag,
+			 int *startp, int *endp, int search_start)
+{
+  int result = 0;
+  int is = search_start, ie;
+  int indicator_present = page_simple_tag_bounds (page, indicator, &is, &ie);
+
+  /* If there is an indicator here, then it *might* belong to TAG.
+     Find out by finding TAG boundaries in the page.  For each
+     boundary found, if the start is past the indicator, then the
+     indicator belongs to us.  Stop when there are no more TAGs found
+     within range, or if the indicator position falls within the start
+     and end of a TAG. */
+  if (indicator_present)
+    {
+      int ts = *startp, te;
+
+      *startp = is;
+      *endp = ie;
+
+      while (1)
+	{
+	  int found = page_complex_tag_bounds (page, tag, &ts, &te);
+
+	  /* If there isn't another complex tag matching TAG, then this
+	     indicator is ours. */
+	  if (found == 0)
+	    {
+	      result = 1;
+	      break;
+	    }
+
+	  /* Found a tag pair.  If the tag start is past the indicator start,
+	     then this indicator is ours. */
+	  if (ts > is)
+	    {
+	      result = 1;
+	      break;
+	    }
+
+	  /* Well, we found a tag pair in the page.  The beginning of
+	     the tag pair is before the indicator.  If the end is before
+	     the indicator as well, then we don't have an answer yet.
+	     Otherwise, the tag doesn't belong to us. */
+	  if (te > ie)
+	    {
+	      result = -1;
+	      break;
+	    }
+
+	  ts = te;
+	}
+    }
+  return (result);
+}
+
 #if defined (NEVER_USED)
 /* Extract the value side of a randomly named variable in PAGE.
    For example, if PAGE contains `<DOCTITLE="this">', then the
@@ -1142,6 +1241,7 @@ page_assigned_label (PAGE *page, char *tag)
 	  if ((page->buffer[i] == delimited) ||
 	      (!delimited &&
 	       ((page->buffer[i] == '>')  ||
+		(page->buffer[i] == RIGHT_BRACKET) ||
 		(whitespace (page->buffer[i])))))
 	    break;
 	}
