@@ -31,6 +31,8 @@
 #define COMPILING_PARSER_C 1
 #include "language.h"
 
+char LEFT_BRACKET = '<';
+char RIGHT_BRACKET = '>';
 
 extern void initialize_external_functions (Package *p);
 
@@ -86,6 +88,12 @@ pagefunc_restore_environment (PageEnv *env)
   free (env);
 }
 
+PAGE *
+parser_top_page (void)
+{
+  return (ThePage);
+}
+
 /* Gets 1 when mhtml::inhibit-comment-parsing has a value, 0 otherwise. */
 int mhtml_inhibit_comment_parsing = 0;
 
@@ -96,6 +104,11 @@ int mhtml_decimal_places = 2;
 /* Gets 1 when mhtml::parser-trace-mode has a value, 0 otherwise. */
 int mhtml_parser_trace_mode = 0;
 
+#if defined (METAHTML_PROFILER)
+/* Gets 1 when mhtml::profile-functions has a value, 0 otherwise. */
+int mhtml_profiling_functions = 0;
+#endif
+
 void
 pagefunc_initialize_notifiers (void)
 {
@@ -107,6 +120,10 @@ pagefunc_initialize_notifiers (void)
   symbol_notify_value (sym, &mhtml_decimal_notify);
   sym = symbol_intern ("mhtml::parser-trace-mode");
   symbol_notify_value (sym, &mhtml_parser_trace_mode);
+#if defined (METAHTML_PROFILER)
+  sym = symbol_intern ("mhtml::profile-functions");
+  symbol_notify_value (sym, &mhtml_profiling_functions);
+#endif /* METAHTML_PROFILER */
 }
 
 static int syntax_checking = 0;
@@ -344,7 +361,12 @@ read_sexp_1 (char *string, int *start, int stop_at_equals_p, int one_list)
       gobbled = 0;
       while (!gobbled)
 	{
-	  register int c = string[i++];
+	  register int orig_c = string[i++];
+	  register int c;
+
+	  c = orig_c;
+	  if (orig_c == LEFT_BRACKET) c = '<';
+	  if (orig_c == RIGHT_BRACKET) c = '>';
 
 	  switch (c)
 	    {
@@ -400,13 +422,13 @@ read_sexp_1 (char *string, int *start, int stop_at_equals_p, int one_list)
 	      break;
 
 	    case '<':
-	      workspace[windex++] = '<';
+	      workspace[windex++] = orig_c;
 	      if (!quoted)
 		depth++;
 	      break;
 
 	    case '>':
-	      workspace[windex++] = '>';
+	      workspace[windex++] = orig_c;
 	      if (!quoted)
 		{
 		  depth--;
@@ -433,6 +455,7 @@ read_sexp_1 (char *string, int *start, int stop_at_equals_p, int one_list)
 		    {
 		      i++;
 		      workspace[windex] = '\0';
+		      gobbled++;
 		    }
 		  else
 		    workspace[windex] = c;
@@ -797,7 +820,7 @@ page_process_page_internal (PAGE *page)
 
       for (i = search_start; i < page->bindex; i++)
 	{
-	  if (page->buffer[i] == '<')
+	  if ((page->buffer[i] == LEFT_BRACKET) || (page->buffer[i] == '<'))
 	    break;
 
 	  /* If there is a semicolon comment here, ignore it now. */
@@ -829,14 +852,14 @@ page_process_page_internal (PAGE *page)
 	  int fname_beg;
 	  int fname_end;
 	  int fname_len;
+	  char openers_opener = page->buffer[i];
 
 	  search_start = i;
 	  fname_beg = ++i;
 
 	  for (; (c = page->buffer[i]) != '\0'; i++)
-	    if ((c == ' ') || (c == '>') ||
-		(c == '\t') || (c == '\r') ||
-		(c == '\n'))
+	    if ((c == ' ') || (c == '>') || (c == RIGHT_BRACKET) ||
+		(c == '\t') || (c == '\r') || (c == '\n'))
 	      break;
 
 	  if (!c)
@@ -916,7 +939,8 @@ page_process_page_internal (PAGE *page)
 		{
 		  /* The MTHML programmer didn't close the opener correctly.
 		     Tell them, and then ignore the text. */
-		  page_debug ("Closing tag missing for <%s ...>", desc->tag);
+		  page_debug ("Closing tag missing for %c%s ...%c",
+			      LEFT_BRACKET, desc->tag, RIGHT_BRACKET);
 		  if (syntax_checking)
 		    {
 		      syntax_failure = 1;
@@ -965,7 +989,8 @@ page_process_page_internal (PAGE *page)
 		      int body_len, body_end;
 		      char *closer = (char *)xmalloc (3 + fname_len);
 
-		      closer[0] = '<';
+		      closer[0] = openers_opener;
+
 		      closer[1] = '/';
 		      /* We'd like to copy the desc->tag, but if this is
 			 a user-function that has been copy-var'd, they
@@ -1030,22 +1055,33 @@ page_process_page_internal (PAGE *page)
 		      if (mhtml_parser_trace_mode || desc->debug_level > 5)
 			{
 			  display_body = strdup (open_body ? open_body : "");
-			  if (strlen (display_body) > 33)
-			    strcpy (display_body + 30, "...");
+			  if (strlen (display_body) > 79)
+			    strcpy (display_body + 76, "...");
 
 			  if (mhtml_parser_trace_mode)
-			    page_debug ("%sEntering %s: <%s %s>",
+			    page_debug ("%sEntering %s: %c%s %s%c",
 					parser_trace_indentation (),
-					desc->tag, desc->tag,
-					mhtml_funargs (vars));
+					desc->tag,
+					LEFT_BRACKET, desc->tag,
+					mhtml_funargs (vars),
+					RIGHT_BRACKET);
 			  else
-			    page_debug ("Entering <%s %s>",
-					desc->tag, display_body);
+			    page_debug ("Entering %c%s %s%c",
+					LEFT_BRACKET, desc->tag,
+					display_body, RIGHT_BRACKET);
 
 			}
 
 		      parser_trace_indentation_counter++;
+		      {
+#if defined (METAHTML_PROFILER)
+			struct timeval start_time = { 0, 0 };
+			struct timeval end_time = { 0, 0 };
 
+			if (mhtml_profiling_functions)
+			  gettimeofday (&start_time, (struct timezone *)NULL);
+#endif
+			
 		      if (uf)
 			mhtml_execute_function
 			  (uf, page, body, vars, start, end, &search_start,
@@ -1054,7 +1090,47 @@ page_process_page_internal (PAGE *page)
 			(*desc->fun)
 			  (page, body, vars, start, end, &search_start,
 			   desc->debug_level);
-
+#if defined (METAHTML_PROFILER)
+		      if (mhtml_profiling_functions &&
+			  ((start_time.tv_usec != 0) &&
+			   (start_time.tv_sec != 0)))
+			{
+			  double usecs_start, usecs_end, elapsed;
+			  gettimeofday (&end_time, (struct timezone *)NULL);
+			  usecs_start = (double)((start_time.tv_sec * 1.0e6)
+						 + start_time.tv_usec);
+			  usecs_end = (double)((end_time.tv_sec * 1.0e6)
+					       + end_time.tv_usec);
+			  elapsed = usecs_end - usecs_start;
+			  
+			  if (uf)
+			    {
+			      if (!uf->profile_info)
+				{
+				  uf->profile_info = (PROFILE_INFO *)
+				    xmalloc (sizeof (PROFILE_INFO));
+				  uf->profile_info->usecs_spent = 0.0;
+				  uf->profile_info->times_called = 0;
+				}
+			      uf->profile_info->times_called++;
+			      uf->profile_info->usecs_spent += elapsed;
+			    }
+			  else
+			    {
+			      if (!desc->profile_info)
+				{
+				  desc->profile_info = (PROFILE_INFO *)
+				    xmalloc (sizeof (PROFILE_INFO));
+				  desc->profile_info->usecs_spent = 0.0;
+				  desc->profile_info->times_called = 0;
+				}
+			      desc->profile_info->times_called++;
+			      desc->profile_info->usecs_spent += elapsed;
+			    }
+			}
+#endif /* METAHTML_PROFILER */
+		      }
+			      
 		      if (search_start < 0)
 			{
 			  page_debug ("PPI: `%s' bashed SEARCH_START!",
@@ -1067,13 +1143,17 @@ page_process_page_internal (PAGE *page)
 		      if (mhtml_parser_trace_mode || desc->debug_level > 5)
 			{
 			  if (mhtml_parser_trace_mode)
-			    page_debug ("%sLeaving %s: <%s %s>",
+			    page_debug ("%sLeaving %s: %c%s %s%c",
 					parser_trace_indentation (),
-					desc->tag, desc->tag,
-					mhtml_funargs (vars));
+					desc->tag,
+					LEFT_BRACKET, desc->tag,
+					mhtml_funargs (vars),
+					RIGHT_BRACKET);
 			  else
-			    page_debug ("Leaving <%s %s>",
-					desc->tag, display_body);
+			    page_debug ("Leaving %c%s %s%c",
+					LEFT_BRACKET,
+					desc->tag, display_body,
+					RIGHT_BRACKET);
 			  free (display_body);
 			}
 
@@ -1197,6 +1277,7 @@ mhtml_add_user_function (int type, char *name, char *body, Package *vars)
       Symbol *sym;
 
       uf = (UserFunction *)xmalloc (sizeof (UserFunction));
+      memset (uf, 0, sizeof (UserFunction));
       uf->type = type;
       uf->debug_level = debug_level ? atoi (debug_level) : 0;
       uf->name = strdup (name);
@@ -1271,6 +1352,9 @@ mhtml_add_user_function (int type, char *name, char *body, Package *vars)
 	  i = 0;
 	  while ((c = b[i]) != '\0')
 	    {
+	      if (c == LEFT_BRACKET) c = '<';
+	      if (c == RIGHT_BRACKET) c = '>';
+
 	      switch (c)
 		{
 		case '"':
@@ -1420,7 +1504,8 @@ mhtml_execute_function (UserFunction *uf, PFunArgs, char *attr)
 
 		      for (k = 1; j < body->bindex; j++)
 			{
-			  if (body->buffer[j] == '"')
+			  if ((body->buffer[j] == '\\') ||
+			      (body->buffer[j] == '"'))
 			    setval[k++] = '\\';
 
 			  setval[k++] = body->buffer[j];
@@ -1485,7 +1570,7 @@ mhtml_execute_function (UserFunction *uf, PFunArgs, char *attr)
 	  if (uf->named_parameters != (char **)NULL)
 	    {
 	      BPRINTF_BUFFER *wrapper = bprintf_create_buffer ();
-	      bprintf (wrapper, "<set-var");
+	      bprintf (wrapper, "%cset-var", LEFT_BRACKET);
 	      for (i = 0; uf->named_parameters[i] != (char *)NULL; i++)
 		{
 		  char *value = get_positional_arg (vars, i);
@@ -1504,7 +1589,7 @@ mhtml_execute_function (UserFunction *uf, PFunArgs, char *attr)
 
 		      for (j = 0, k = 1; value[j] != '\0'; j++)
 			{
-			  if (value[j] == '"')
+			  if ((value[j] == '"') || (value[j] == '\\'))
 			    setval[k++] = '\\';
 
 			  setval[k++] = value[j];
@@ -1522,19 +1607,52 @@ mhtml_execute_function (UserFunction *uf, PFunArgs, char *attr)
 		      free (setval);
 		    }
 		}
-	      bprintf (wrapper, ">");
+	      bprintf (wrapper, "%c", RIGHT_BRACKET);
 	      parameter_setter = wrapper->buffer;
 	      free (wrapper);
 	    }
 
 	  if (packname)
-	    bprintf_insert (page, start, "<in-package %s>%s%s</in-package>",
-			    packname, parameter_setter ? parameter_setter : "",
-			    subber->buffer);
+	    {
+	      /* Handle "local" package inline. */
+	      if (strcmp (packname, "local") == 0)
+		{
+		  Package *current_package = CurrentPackage;
+		  char *x;
+
+		  CurrentPackage = symbol_get_package ((char *)NULL);
+
+		  if (parameter_setter)
+		    {
+		      x = mhtml_evaluate_string (parameter_setter);
+		      xfree (x);
+		    }
+
+		  x = mhtml_evaluate_string (subber->buffer);
+		  if (!empty_string_p (x))
+		    {
+		      bprintf_insert (page, start, "%s", x);
+		      *newstart += strlen (x);
+		    }
+		  xfree (x);
+		  CurrentPackage = current_package;
+		}
+	      else
+		{
+		  bprintf_insert (page, start,
+				  "%cin-package %s%c%s%s%c/in-package%c",
+				  LEFT_BRACKET, packname, RIGHT_BRACKET,
+				  parameter_setter ? parameter_setter : "",
+				  subber->buffer,
+				  LEFT_BRACKET, RIGHT_BRACKET);
+		}
+	    }
 	  else
-	    bprintf_insert (page, start, "%s%s", 
-			    parameter_setter ? parameter_setter : "",
-			    subber->buffer);
+	    {
+	      bprintf_insert (page, start, "%s%s",
+			      parameter_setter ? parameter_setter : "",
+			      subber->buffer);
+	    }
 	}
 
       page_free_page (subber);
@@ -1699,6 +1817,34 @@ mhtml_canonicalize_file_name (char *input, char *docroot, char *relpref)
   return (result);
 }
 
+char *
+mhtml_canonicalize_file_name_argument (char *argument)
+{
+  char *temp = mhtml_evaluate_string (argument);
+  char *result = (char *)NULL;
+
+  if (!empty_string_p (temp))
+    {
+      if (pagefunc_get_variable ("isp::web-relative-pathnames"))
+	{
+	  char *incpref = pagefunc_get_variable ("%%::incpref");
+	  char *relpref = pagefunc_get_variable ("%%::relpref");
+
+	  if (!incpref)
+	    incpref = pagefunc_get_variable ("mhtml::include-prefix");
+	  if (!relpref)
+	    relpref = pagefunc_get_variable ("mhtml::relative-prefix");
+
+	  result = mhtml_canonicalize_file_name (temp, incpref, relpref);
+	}
+      else
+	result = temp;
+    }
+
+  if (result != temp) xfree (temp);
+  return (result);
+}
+
 /* Set the debugging level for the function named in SYM to
    be the value of SYM. */
 void
