@@ -120,6 +120,9 @@ session_begin (char *key, int allow_multiple)
       session_put_info_internal (info);
       session_free (info);
     }
+
+  /* New sessions cause the reaping of old ones. */
+  session_reap ();
   return (sid);
 }
 
