@@ -62,7 +62,7 @@ sd_info_to_package (SESSION_INFO *info, Package *package)
 	  pair = CAR (list);
 	  list = CDR (list);
 
-	  if (CONS_P (pair) & STRING_P (CAR (pair)))
+	  if (CONS_P (pair) && STRING_P (CAR (pair)))
 	    {
 	      char *tag;
 
