@@ -191,7 +191,8 @@ static void
 pf_with_open_stream (PFunArgs)
 {
   char *varname = mhtml_evaluate_string (get_positional_arg (vars, 0));
-  char *filename = mhtml_evaluate_string (get_positional_arg (vars, 1));
+  char *filename = mhtml_canonicalize_file_name_argument
+    (get_positional_arg (vars, 1));
   int jump_again = 0;
 
   jump_again = 0;
@@ -206,9 +207,9 @@ pf_with_open_stream (PFunArgs)
       int flags = notimeout ? stream_NOTIMEOUT : 0;
       int stream_fd = -1;
 
-      if (modename) free (modename);
-      if (typename) free (typename);
-      if (notimeout) free (notimeout);
+      xfree (modename);
+      xfree (typename);
+      xfree (notimeout);
 
       switch (type)
 	{
@@ -284,12 +285,12 @@ pf_with_open_stream (PFunArgs)
 	  page_syserr ("<with-open-stream%s%s>: %s",
 		       temp ? " " : "", temp ? temp : "",
 		       (char *)strerror (errno));
-	  if (temp) free (temp);
+	  xfree (temp);
 	}
     }
 
-  if (filename) free (filename);
-  if (varname) free (varname);
+  xfree (filename);
+  xfree (varname);
   if (jump_again) longjmp (page_jmp_buffer, 1);
 }
 
@@ -522,11 +523,13 @@ stream_read (Stream *stream, char char_to_stop_at)
 
       memmove (buffer + buffer_index, minbuf, mindex);
       buffer_index += mindex;
-      buffer[buffer_index] = '\0';
     }
 
   stream_last_read_object_size = buffer_index;
 
+  if (buffer != (char *)NULL)
+    buffer[buffer_index] = '\0';
+
   return (buffer);
 }
 
@@ -639,8 +642,7 @@ pf_stream_put_contents (PFunArgs)
 	    }
 	}
 
-      if (name != (char *)NULL)
-	free (name);
+      xfree (name);
     }
 }
 
@@ -667,7 +669,7 @@ pf_stream_get (PFunArgs)
       free (result);
     }
 
-  if (stop_at != (char *)NULL) free (stop_at);
+  xfree (stop_at);
 }
 
 static void
@@ -721,9 +723,9 @@ pf_stream_get_contents (PFunArgs)
       free (result);
     }
 
-  if (name) free (name);
-  if (chunk_size_arg) free (chunk_size_arg);
-  if (stop_at) free (stop_at);
+  xfree (name);
+  xfree (chunk_size_arg);
+  xfree (stop_at);
 }
 
 static void
