@@ -461,6 +461,7 @@ symbol_intern_in_package (Package *package, char *name)
       SymbolList *new = (SymbolList *)xmalloc (sizeof (SymbolList));
 
       symbol = (Symbol *)xmalloc (sizeof (Symbol));
+      memset (symbol, 0, sizeof (Symbol));
       symbol->flags = 0;
       symbol->name_len = strlen (name);
       symbol->name = (char *)xmalloc (1 + symbol->name_len);
@@ -759,12 +760,14 @@ symbol_free (Symbol *symbol)
 	{
 	case symtype_STRING:
 	  free (symbol->name);
+	  if (symbol->preserved_name) free (symbol->preserved_name);
 	  free_array (symbol->values);
 	  free (symbol);
 	  break;
 
 	case symtype_BINARY:
 	  free (symbol->name);
+	  if (symbol->preserved_name) free (symbol->preserved_name);
 	  datablock_free ((Datablock *)symbol->values);
 	  free (symbol);
 	  break;
@@ -888,7 +891,9 @@ symbol_rename (Symbol *symbol, char *newname)
   newsym->values_index = sym->values_index;
   newsym->values_slots = sym->values_slots;
   newsym->type = sym->type;
+  newsym->flags = sym->flags;
   free (sym->name);
+  if (sym->preserved_name) free (sym->preserved_name);
   free (sym);
 
   return (newsym);
@@ -908,6 +913,7 @@ symbol_move (Symbol *symbol, Package *package)
   symbol_free (symbol_remove_in_package (package, symbol->name));
   list->next = package->table->symbol_list[row];
   package->table->symbol_list[row] = list;
+  package->table->entries++;
   list->symbol = symbol;
   symbol->package = (void *)package;
 
@@ -960,10 +966,18 @@ symbol_full_name (Symbol *symbol)
       strcpy (name, pack->name);
       name[pack->name_len + 0] = ':';
       name[pack->name_len + 1] = ':';
-      strcpy (name + pack->name_len + 2, symbol->name);
+      if (symbol->preserved_name)
+	strcpy (name + pack->name_len + 2, symbol->preserved_name);
+      else
+	strcpy (name + pack->name_len + 2, symbol->name);
     }
   else
-    name = strdup (symbol->name);
+    {
+      if (symbol->preserved_name)
+	name = strdup (symbol->preserved_name);
+      else
+	name = strdup (symbol->name);
+    }
 
   return (name);
 }
