@@ -33,10 +33,13 @@
 
 #define DEFAULT_PACKAGE_NAME "DEFAULT"
 
-/* What a symbol looks like. */
+/* What a symbol looks like.
+   Note that PRESERVED_NAME is almost always NULL, and can only be used
+   when it has been explicity set. */
 typedef struct
 {
   char *name;			/* The characters of the symbol name. */
+  char *preserved_name;		/* The symbol name, with case preserved. */
   int name_len;			/* The length of the symbol name. */
   char **values;		/* The values associated with this symbol. */
   int values_index;		/* Number of values stored within. */
