@@ -34,3 +34,25 @@ extern void symbol_dump_package (int fd, Package *package);
 
 /* Read a package from FD, installing the symbols and values found within. */
 extern Package *symbol_load_package (int fd);
+
+/* Return an SD (like an FD, but for the string buffer pool.
+   Newly allocates a buffer.
+   The SD returned is suitable for passing to symbol_load_package ()
+   and symbol_dump_package (). */
+extern int symdump_open_string_data (void);
+
+/* Return the BPRINTF_BUFFER corresponding to SD, and return the allocated
+   slot back to the pool.  Does NOT free the BPRINTF_BUFFER. */
+extern BPRINTF_BUFFER *symdump_close_string_data (int sd);
+
+/* Read LEN bytes from SD into DATA. */
+extern int symdump_read_string_data (int sd, int len, unsigned char *data);
+
+/* Write LEN bytes to SD from DATA. */
+extern int symdump_write_string_data (int sd, int len, unsigned char *data);
+
+/* Set the current offstring in SD to OFFSET. */
+extern void symdump_seek_string_data (int sd, int offset);
+
+/* Set the current offstring in SD to OFFSET. */
+extern void symdump_set_string_data_buffer_size (int sd, int end);
