@@ -238,7 +238,7 @@ mhttpd_date_format (long ticks)
     hours_from_gmt = ((tz.tz_minuteswest / 60) / 60);
   }
 #else
-#  if defined (__linux__) || defined (Solaris) || defined (sgi) || defined (_AIX)
+#  if defined (__linux__) || defined (Solaris) || defined (sgi) || defined (_AIX) || defined (_SCO_DS)
   hours_from_gmt = (timezone / 60) / 60;
 #  else
   hours_from_gmt = (date->tm_gmtoff / 60) / 60;
