@@ -970,7 +970,16 @@ mdb_breakpoint (MDBArgs)
     }
   return (result);
 }
-      
+
+#if defined (MDB_EDITOR_COMMAND)
+static char *
+mdb_edit (MDBArgs)
+{
+  int line_number, defaulted;
+  char *filename;
+  char *result = (char *)NULL;
+}
+#endif
 
 static int include_recursive_calls = 0;
 
