@@ -59,7 +59,10 @@ extern char *rl_completer_word_break_characters;
 extern char *rl_line_buffer;
 extern char **completion_matches (char *, CPFunction *);
 extern CPPFunction *rl_attempted_completion_function;
-
+typedef int Function (void);
+extern int rl_insert_close (int count, int key);
+extern int rl_bind_key (int key, Function *);
+;
 /* Helper functions for Readline. */
 static char **attempt_mdb_completion (char *text, int start, int end);
 static char *mdb_command_completion_function (char *text, int state);
@@ -94,7 +97,9 @@ mdb_initialize_readline (void)
 
   /* Need to modify this from the default; `$', `{', `\', and ``' are not
      word break characters. */
-  rl_completer_word_break_characters = " \t\n\"'@><=;|&("; /**/
+  rl_completer_word_break_characters = " \t\n\"'@><=;|&(";
+
+  rl_bind_key ('>', (Function *)rl_insert_close);
 
   mdb_readline_initialized = 1;
 }
@@ -209,12 +214,12 @@ pagefunc_completion_function (char *hint_text, int state)
       hint_len = strlen (hint);
       local_index = 0;
 
-      /* Skip any number of open braces. */
+      /* Skip any number of open braces and slash characters. */
       {
 	register int i;
 
 	for (i = 0; hint[i]; i++)
-	  if (hint[i] != '<')
+	  if ((hint[i] != '<') && (hint[i] != '/'))
 	    break;
 
 	if (i)
