@@ -22,6 +22,10 @@
    675 Mass Ave, Cambridge, MA 02139, USA. */
 #define READLINE_LIBRARY
 
+#if defined (HAVE_CONFIG_H)
+#  include <config.h>
+#endif
+
 /* **************************************************************** */
 /*								    */
 /*			VI Emulation Mode			    */
