@@ -50,10 +50,6 @@ static void usage (void);
 static char *progname = (char *)NULL;
 static PAGE *input_contents = (PAGE *)NULL;
 
-#if defined (CHECK_TIMESTAMP)
-static void check_timestamp (char *filename);
-#endif
-
 extern char *metahtml_copyright_string;
 
 /* This program takes a single filename as an argument, and runs the
@@ -72,10 +68,6 @@ main (int argc, char *argv[])
 
   parse_program_args (argc, argv);
 
-#if defined (CHECK_TIMESTAMP)
-  check_timestamp (argv[0]);
-#endif
-
   if (!input_contents)
     usage ();
 
@@ -108,6 +100,8 @@ prep_for_page_process (PAGE *page, int argc, char *argv[])
 {
   register int i;
 
+  bootstrap_metahtml (0);
+
   if (page && page->buffer)
     {
       if (page->buffer[0] == '#' && page->buffer[1] == '!')
@@ -185,26 +179,23 @@ parse_program_args (int argc, char *argv[])
 
 	  pagefunc_set_variable (name, value);
 	}
-      else if (strncmp (arg, "--", 2) != 0)
+      else if (!filename && (*arg != "-"))
+	filename = arg;
+      else if (filename)
 	{
-	  if (filename)
-	    {
-	      static Symbol *sym = (Symbol *)NULL;
+	  static Symbol *sym = (Symbol *)NULL;
 
-	      if (!sym) sym = symbol_intern ("mhtml::program-arguments");
+	  if (!sym) sym = symbol_intern ("mhtml::program-arguments");
 
-	      symbol_add_value (sym, arg);
-	    }
-	  else
-	    filename = arg;
+	  symbol_add_value (sym, arg);
 	}
       else
 	usage ();
     }
 
-  if (!filename && !input_contents)
+  if (!filename)
     usage ();
-  else if (filename)
+  else
     input_contents = page_read_template (filename);
 
   {
@@ -223,51 +214,11 @@ parse_program_args (int argc, char *argv[])
 	  {
 	    char dir[1024];
 	    temp = getcwd (dir, sizeof (dir));
-	    temp = strrchr (dir, '/');
 	    if (temp != (char *)NULL)
-	      {
-		*temp = '\0';
-		pagefunc_set_variable ("mhtml::include-prefix", dir);
-	      }
+	      pagefunc_set_variable ("mhtml::include-prefix", dir);
+	    else
+	      pagefunc_set_variable ("mhtml::include-prefix", "/");
 	  }
       }
-
-    /* Define default functions here. */
-    {
-      PAGE *p = page_create_page ();
-
-      bprintf (p, "<defsubst meta-html>");
-      bprintf (p, "<font color=\"#00309B\">Meta-</font>HTML");
-      bprintf (p, "<if <not <get-var mhtml::printed-tm>>");
-      bprintf (p, "<prog <sup><font size=-2>TM</font></sup>");
-      bprintf (p, "<increment mhtml::printed-tm>>>");
-      bprintf (p, "</defsubst>");
-      bprintf (p, "<defmacro ptext>");
-      bprintf (p, "<plain-text first-char=<font size=+2> nobr=true>");
-      bprintf (p, "%s", "%body</plain-text></defmacro>");
-      page_process_page (p);
-      page_free_page (p);
-    }
   }
 }
-
-#if defined (CHECK_TIMESTAMP)
-#include "../libtimestamp/timestamp.h"
-
-static void
-check_timestamp (char *filename)
-{
-  if (check_stamp_expired (filename))
-    {
-      fprintf (stderr, "Sorry, but your 45 day free trial has ended.\n");
-      fprintf (stderr, "You may obtain another copy of MHC from\n");
-      fprintf (stderr, "\n\thttp://www.metahtml.com\n\n");
-      fprintf (stderr, "We hope that you have enjoyed the use of our\n");
-      fprintf (stderr, "Meta-HTML Standalone Processor, and will continue\n");
-      fprintf (stderr, "to watch our site for new additions to our product\n");
-      fprintf (stderr, "suite.\n");
-      exit (2);
-    }
-}
-#endif /* CHECK_TIMESTAMP */
-
